package bank3tier.db;//DbObject.java

/*
 * 2005-04-05  Knabe  Operative persistence operations moved to interface Pool, class _JoraPool.
 * 2002-07-08  Knabe  oid state evolves from isCreated() to isPersistent to isDeleted
 * 2002-04-30  Knabe  dbUpdate() checks new db__updateVersion counter.
 * 2002-04-11  Knabe  All modifying operation names begin with prefix db
 * 2001-05-28  Knabe  Preventing duplicate insert.
 *   All mutator operations only as protected, against usage from UI tier.
 * 2001-05-03  Knabe  OID type changed from int to Integer
 * 2001-04-22  Knabe  All object related DB operations here as protected
 * Created on 28. Februar 2001, 15:31
 */



/**Base class for all objects storable in a persistency pool. This class is purely generic.
  All application specific code (especially table creation) is collected
  in the class {@link DbApplication}.
  The persistence layer is technically implemented in a subclass {@link Pool}. 
  Each object is identified by an OID. The OID (Object IDentification) is a
  number unique for each object in our application.
  This class supports the OR-Mapping according to Scott Ambler. But his
  retrieve-Operation is not implemented, as it is only necessary for
  efficiency optimization.
  The state of a DbObject evolves from isCreated() by Pool.save() to
  isPersistent() and by Pool.delete() to isDeleted().
*/
public abstract class DbObject {

    public static final String oidSuffix = "_oid";

    /**Unique Object IDentification for each DbObject instance*/
    /*package*/ Integer oid = Pool.createdOid;

    /**Modification version counter for this object, shall be incremented
      by any insert oder update operation.
    */
    /*package*/ int db__updateVersion = 0;
    
    /**Der Pool, zu dem dieses Objekt gehrt. Sollte nicht null sein, um Pool-Operationen 
     * darauf ausfhren zu knnen. Diese Pool-Referenz ist transient, damit sie nicht
     * in der Datenbank gespeichert wird. Stattdessen wird beim Laden eines persistenten Objekts
     * aus der Datenbank das Attribut _pool initialisiert.
     * Diese Pool-Referenz ist fr eine Pool-basierte Persistenz nicht unbedingt 
     * ntig, macht aber die Handhabung bequemer.
     **/
    private transient Pool _pool;


    /**Returns the OID of this object.*/
    public final Integer getOid(){return oid;}

    /**Returns true, if this object has been newly created without being made persistent*/
    public final boolean isCreated(){return oid==Pool.createdOid;}

    /**Returns true, if this object has been made persistent*/
    public final boolean isPersistent(){return oid!=null && oid.intValue() > 0;}

    /**Returns true, if this object has been deleted from the persistence mechanism*/
    public final boolean isDeleted(){return oid==Pool.deletedOid;}

    /**Initializes the Pool, to which this object belongs to. Must be called only by the persistence manager.*/
    /*package*/ void initPool(final Pool io_pool) throws OwnerPoolChangedExc {
        if(_pool!=null){
            throw new OwnerPoolChangedExc(this);
        }
        _pool = io_pool;
    }
    
    /*Creates a DbObject, not yet belonging to any Pool.*/
    public DbObject(){}
    
    /**Creates a DbObject belonging to this Pool. Only for testing purposes.*/
    protected DbObject(final Pool io_pool){
        _pool = io_pool;
    }
    
    /**Returns the Pool, to which this object belongs to. Should be called only by the persistence manager.*/
    protected Pool getPool(){
        return _pool;
    }

    /**Must not change the owning Pool of object*/
    public static class OwnerPoolChangedExc extends multex.Exc {
        public OwnerPoolChangedExc(final DbObject i_object){
            super("Must not change the owning Pool of object with OID {0} of class {1}"
                , i_object.getOid(), i_object.getClass().getName()
            );
        }
    }


}