package bank3tier.db; //Pool

import java.util.List;


/**Operative interface to the persistence layer similar to Siedersleben, page 204, 
 * but with some more operations. 
 * This is the interface, which should be used mostly by the business logic layer.
 * @author Christoph Knabe 2005-03-28
 */
public interface Pool {

	/*package*/ static final Integer createdOid = null;
    /*package*/ static final Integer deletedOid = new Integer(0);
    
    /**Saves the actual state of the object into the pool.
     * If the object was not persistent before, it gets a new OID.
     * Its Pool is set to this Pool.
     * @return true newly inserted into the pool.
     * @throws ConcurrentUpdateExc  Another user modified this object in the Pool before you.
     *                              Thus you should repeat your transaction.
     */
    boolean save(DbObject io_object) throws ConcurrentUpdateExc, DbObject.OwnerPoolChangedExc;
    
    /**Deletes the object from the pool.
     * Additonally mark the object as not valid by giving it the OID deletedOid.
     * @return true it was in the pool.
     * TODO May be ConcurrentUpdateExc, too, if it was updated by another user
     * after loading.
     */
    boolean delete(DbObject io_object);
    
    /**Returns the element of the Pool with the given Class and OID.*/
    DbObject find(Class i_persistentClass, Integer i_oid) throws NoPersistentClassExc, OidNotFoundExc;
    
    /**Returns a List with all elements of the given Class in the Pool.*/
    /*DbObject*/List findAll(Class i_persistentClass) throws NoPersistentClassExc;
    
    /** Returns the one unique object by a "query by example"
     *  from the Pool.
     *  The returned object will be dynamically an instance of the same type
     *  as the actual example object.
     *  Note: The selected object should
     *  have no components of InputStream, Blob or Clob type, because
     *  their data will be not available after closing the Cursor used to get
     *  it.
     */
    DbObject findUniqueByExample(DbObject i_object) throws ExampleNotFoundExc, ExampleNotUniqueExc;

    /**Returns all DbObject-s, which have in the 
     * non-null, non-transient attributes an equal value as i_object.
     * The returned objects will be dynamically an instance of the same type
     * as the actual example object.
     * The resultant List may be empty.
     */ 
    public /*DbObject*/List findManyByExample(DbObject i_object);

    /**Returns true, if there is an object of i_linkClass, which is a link between i_left and i_right.
     * @throws NoPersistentClassExc i_linkClass is not a persistent class.*/
    public boolean areLinked(DbObject i_left, Class i_linkClass, DbObject i_right) throws NoPersistentClassExc;

    /**Returns the DbObject of the class i_partnerClass, which is linked
      to from i_object. The name of the class i_partnerClass is used
      to determine a) the Table, in which to search, and b) the value of the
      OID looked for is taken from the attribute of i_object, whose name is formed
      as a concatenation of the partner class name, and the suffix oidSuffix.
     * @throws NoPersistentClassExc i_partnerClass or the Class of i_object is not a subclass of bank3tier.db.DbObject
    */
    public DbObject getLink(DbObject i_object, Class i_partnerClass) throws LinkFailure, NoPersistentClassExc;
    
    /**Returns all objects of class i_resultClass from the Pool, 
     * which fulfill the query i_queryClass with the given arguments.
     * @throws NoPersistentClassExc i_resultClass is not a subclass of bank3tier.db.DbObject
     * @throws NoQueryClassExc i_queryClass is not a subclass of bank3tier.db.Query
     */
    /*DbObject*/List findManyByQuery(Class i_resultClass, Class i_queryClass, Object[] i_args) 
    throws NoPersistentClassExc, NoQueryClassExc, ArgumentCountExc, ArgumentTypeExc;

    /**Object with OID {0} not found in table \"{1}\"*/
    public static class OidNotFoundExc extends multex.Exc {
        public OidNotFoundExc(final Integer i_oid, final jora.Table i_table){
            super("Object with OID {0} not found in table \"{1}\"",
                i_oid, i_table.name()
            );
        }
    }
    
    /**Object has been modified by another user. You should repeat the transaction.*/
    public static class ConcurrentUpdateExc extends multex.Exc {
        public ConcurrentUpdateExc(final String i_tableName, final Integer i_oid){
            super("{0}-object with OID {1} has been modified by another user. "
                + " You should repeat the transaction", i_tableName, i_oid
            );
        }
    }

    /**Class is not a subclass of bank3tier.db.DbObject. Operation impossible.*/
    public static class NoPersistentClassExc extends multex.Exc {
        public NoPersistentClassExc(final Class i_class){
            super("Class {0} is not a persistent class. Operation impossible."
                , i_class.getName()
            );
        }
    }

    /**No object found matching the example object*/
    public static class ExampleNotFoundExc extends multex.Exc {
        public ExampleNotFoundExc(final DbObject i_object){
            super("No object found matching the example object \"{0}\"", i_object);
        }
    }

    /**Query result not unique with example object*/
    public static class ExampleNotUniqueExc extends multex.Exc {
        public ExampleNotUniqueExc(final DbObject i_object){
            super("Query result not unique with example object \"{0}\"", i_object);
        }
    }

    /**Coming from class \"{0}\" with field \"{1}\" cannot find in partner table \"{2}\" the linked Object with OID {3}*/
    public static class LinkFailure extends multex.Failure {
        public LinkFailure(final Exception i_cause, final String i_ownClassName, final String i_partnerOidFieldName, final String i_partnerTableName, final Integer i_partnerOid){
            super("Coming from class \"{0}\" with field \"{1}\" cannot find in partner table \"{2}\" the linked Object with OID {3} ",
                i_cause, i_ownClassName, i_partnerOidFieldName, i_partnerTableName, i_partnerOid
            );
        }
    }
    
    /**Query Class is not a subclass of query base class.*/
    public static class NoQueryClassExc extends multex.Exc {
        public NoQueryClassExc(final Class i_queryClass, final Class i_baseClass){
            super("Query Class {0} is not a subclass of query base class {1}."
                , i_queryClass.getName(), i_baseClass.getName()
            );
        }
    }
    
    /**Query {0} has been called with {1} arguments, but needs {2}.*/
    public static class ArgumentCountExc extends multex.Exc {
        public ArgumentCountExc(final Class i_queryClass, final int i_providedCount, final int i_neededCount){
            super("Query {0} has been called with {1} arguments, but needs {2}."
                , i_queryClass.getName(), new Integer(i_providedCount), new Integer(i_neededCount)
            );
        }
    }
    
    /**Of query {0} the argument index {1} is of class {2}, but should be of class {3}.*/
    public static class ArgumentTypeExc extends multex.Exc {
        public ArgumentTypeExc(final Class i_queryClass, final int i_argumentIndex, final Class i_argumentClass, final Class i_expectedClass){
            super("Of query {0} the argument index {1} is of class {2}, but should be of class {3}."
                , i_queryClass.getName(), new Integer(i_argumentIndex), i_argumentClass.getName(), i_expectedClass.getName()
            );
        }
    }

}
