//-< Cursor.java >---------------------------------------------------*--------*
// JORA                       Version 2.0        (c) 1998  GARRET    *     ?  *
// (Java Object Relational Adapter)                                  *   /\|  *
//                                                                   *  /  \  *
//                          Created:     10-Jun-98    K.A. Knizhnik  * / [] \ *
//                          Last update: 20-Jun-98    K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Cursor for navigation thru result of SELECT statement
//-------------------------------------------------------------------*--------*

package jora;

import java.util.*;
import java.lang.reflect.*;
import java.sql.*;

/** Cursor is used for successive access to records fetched by SELECT
 *  statement. As far as records can be retrived from several derived tables
 *  (polymorphic form of select), this class can issue several requests
 *  to database. Cursor also provides methods for updating/deleting current
 *  record.
 * @version 2002-06-24 Added checks for SQLWarning
 */
public class Cursor {
    /**
     * A cursor is initially positioned before its first row; the
     * first call to next makes the first row the current row; the
     * second call makes the second row the current row, etc.
     *
     * <P>If an input stream from the previous row is open, it is
     * implicitly closed. The ResultSet's warning chain is cleared
     * when a new row is read.
     *
     * @return object constructed from fetched record or null if there
     * are no more rows
     */
    public Object next() throws Cursor.NextFailure {
        try {
            do {
                if (resultSet == null) {
                    if (table.isAbstract) {
                        table = table.derived;
                        continue;
                    }
                    if (qbeObject != null) {
                        String sql = "select " + table.listOfFields
                        + " from " + table.name() + " " + query;
                        Failure.log("executeQuery as PreparedStatement", sql); //Knabe 01-02-15
                        //final DiagnosticPreparedStatement qbeStmt;
                        synchronized(session.preparedStmtHash) {
                            final Object s = session.preparedStmtHash.get(sql);
                            if (s == null) {
                                qbeStmt = new DiagnosticPreparedStatement
                                (session.connection, sql);
                                session.preparedStmtHash.put(sql, qbeStmt);
                            } else {
                                qbeStmt = (DiagnosticPreparedStatement)s;
                            }
                        }
                        synchronized(qbeStmt) {
                            /*
                            table.bindQueryVariables(qbeStmt, qbeObject);
                            result = qbeStmt.executeQuery();
                            qbeStmt.clearParameters();
                            Retain parameters for diagnostics. Knabe 2001-06-15:
                            */
                            qbeStmt.clearParameters();
                            table.bindQueryVariables(qbeStmt, qbeObject);
                            resultSet = qbeStmt.executeQuery();
                        }
                    } else {
                        if (stmt == null) {
                            stmt = session.connection.createStatement();
                            SqlWarningFailure.check(session.connection);
                        }
                        String sql = "select " + table.listOfFields
                        + " from " + table.name() + " " + query;
                        Failure.log("executeQuery", sql); //Knabe 01-02-15
                    try{resultSet = stmt.executeQuery(sql);}
                    catch(SQLException ex){throw new SqlFailure(new JdbcFailure(ex),sql);}
                    }
                }
                if (resultSet.next()) {
                    SqlWarningFailure.check(resultSet);
                    return currObject = table.load(resultSet);
                }
                resultSet.close();
                resultSet = null;
                currObject = null;
                table = table.derived;
            } while (--nTables != 0);

            if (stmt != null) {
                stmt.close();
            }
        }
        catch (java.sql.SQLException ex) { throw new Cursor.NextFailure(new JdbcFailure(ex)); }
        catch (Exception ex) { throw new Cursor.NextFailure(ex); }
        return null;
    }

    /** Update current record pointed by cursor. This method can be called
     *  only after next() method, which returns non-null object. This objects
     *  is used to update current record fields.<P>
     *
     *  If you are going to update or delete selected records, you should add
     *  "for update" clause to select statement. So parameter of
     *  <CODE>jora.Table.select()</CODE> statement should contain "for update"
     *  clause:
     *  <CODE>record.table.Select("where name='xyz' for update");</CODE><P>
     *
     *  <I><B>Attention!</I></B>
     *  Not all database drivers support update operation with
     *  cursor. This method will not work with such database drivers.
     */
    public void update() throws NoCurrentObjectExc, UpdateFailure {
        if (currObject == null) {
            throw new NoCurrentObjectExc();
        }
        try {
            table.updateVariables(resultSet, currObject);
        }
        catch (SQLException ex) { throw new UpdateFailure(new JdbcFailure(ex)); }
        catch (Exception ex) { throw new UpdateFailure(ex); }
    }

    /** Delete current record pointed by cursor. This method can be called
     *  only after next() method, which returns non-null object.<P>
     *
     *  If you are going to update or delete selected records, you should add
     *  "for update" clause to select statement. So parameter of
     *  <CODE>jora.Table.select()</CODE> statement should contain "for update"
     *  clause:
     *  <CODE>record.table.Select("where name='xyz' for update");</CODE><P>
     *
     *  <I><B>Attention!</I></B>
     *  Not all database drivers support delete operation with cursor.
     *  This method will not work with such database drivers.
     */
    public void delete() throws NoCurrentObjectExc, DeleteFailure {
        if (currObject == null) {
            throw new NoCurrentObjectExc();
        }
        try {
            resultSet.deleteRow();
        }
        catch (SQLException ex) { throw new DeleteFailure(new JdbcFailure(ex)); }
        SqlWarningFailure.check(resultSet);
    }

    /** Extracts no more than <I>maxElements</I> records from database and
     *  store them into array. It is possible to extract rest records
     *  by successive next() or toArray() calls. Selected objects should
     *  have now components of InputStream, Blob or Clob type, because
     *  their data will be not available after fetching next record.
     *
     * @param maxElements limitation for result array size (and also for number
     *  of fetched records)
     * @return Array with objects constructed from fetched records.
     */
    public Object[] toArray(int maxElements) {
        Vector v = new Vector(maxElements < 100 ? maxElements : 100);
        Object o;
        while (--maxElements >= 0 && (o = next()) != null) {
            v.addElement(o);
        }
        Object[] a = new Object[v.size()];
        v.copyInto(a);
        return a;
    }

    /** Store all objects returned by SELECT query into array of Object.
     *  Selected objects should have now components of InputStream, Blob or
     *  Clob type, because their data will be not available after fetching
     *  next record.
     *
     * @return Array with objects constructed from fetched records.
     */
public Object[] toArray() { return toArray(Integer.MAX_VALUE); }

    /** Extracts no more than <I>maxElements</I> records from database and
     *  store them into array. It is possible to extract rest records
     *  by successive next() or toArray() calls. Selected objects should
     *  have now components of InputStream, Blob or Clob type, because
     *  their data will be not available after fetching next record.
     *
     * @param maxElements limitation for result array size (and also for number
     *  of fetched records)
     * @return List with objects constructed from fetched records.
     */
    public List toArrayList(int maxElements) {
        ArrayList al = new ArrayList(maxElements < 100 ? maxElements : 100);
        Object o;
        while (--maxElements >= 0 && (o = next()) != null) {
            al.add(o);
        }
        return al;
    }

    /** Store all objects returned by SELECT query into a list of Object.
     *  Selected objects should have now components of InputStream, Blob or
     *  Clob type, because their data will be not available after fetching
     *  next record.
     *
     * @return Array with objects constructed from fetched records.
     */
public List toArrayList() { return toArrayList(Integer.MAX_VALUE); }

    // Internals

    protected Cursor(Table table, Session session, int nTables, String query) {
        if (session == null) {
            session = ((SessionThread)Thread.currentThread()).session;
        }
        this.table = table;
        this.session = session;
        this.nTables = nTables;
        this.query = query;
    }

    protected Cursor(Table table, Session session, int nTables, Object obj) {
        if (session == null) {
            session = ((SessionThread)Thread.currentThread()).session;
        }
        this.table = table;
        this.session = session;
        this.nTables = nTables;
        qbeObject = obj;
        query = table.buildQueryList(obj);
        stmt = null;
    }

    /**Releases this Cursor object's JDBC resources immediately instead of
     *waiting for this to happen when it is automatically closed by the
     *garbage collection.
     *@author Knabe 2001-02-16
     */
    //2001-05-06  Knabe  SQLException wrapped in Failure exception
    public void close()throws CloseFailure {
        final String tableName = table.name();
        try {
            table = null;
            session = null;
            nTables = 0;
            if(resultSet!=null){resultSet.close();   resultSet = null;}
            query = null;
            if(stmt!=null){stmt.close();   stmt = null;}
            currObject = null;
            qbeObject  = null;
        }
        catch(SQLException ex){throw new CloseFailure(new JdbcFailure(ex), tableName);}
        catch(Exception ex){throw new CloseFailure(ex, tableName);}
    }

    //2001-04-19  Christoph Knabe  With unnamed exception parameters
    //2001-02-27  Christoph Knabe  Creation
    /** Unable to locate next object from Cursor.*/
    public class NextFailure extends jora.Failure {
        public NextFailure(final Exception i_cause){
            super("Unable to locate next object from Cursor on table {0} with query {1}",
              i_cause, table.name(), query
            );
        }
    }

    //2001-04-19  Christoph Knabe  With unnamed exception parameters
    //2001-03-08  Christoph Knabe  Creation
    /** Unable to update actual row of Cursor.
    */
    public class UpdateFailure extends jora.Failure {
        public UpdateFailure(final Exception cause){
            super("Unable to update actual row of cursor for table {0}",
                cause, table.name()
            );
        }
    }

    //2001-04-19  Christoph Knabe  With unnamed exception parameters
    //2001-03-08  Christoph Knabe  Creation
    /** Unable to delete actual row of cursor
    */
    public class DeleteFailure extends jora.Failure {
        public DeleteFailure(final Exception cause){
            super(
                "Unable to delete actual row of cursor for table {0}",
                cause, table.name()
            );
        }
    }

    //2001-05-06  Christoph Knabe  Creation
    /** Unable to delete actual row of cursor
    */
    public class CloseFailure extends jora.Failure {
        public CloseFailure(final Exception cause, final String i_tableName){
            super(
                "Failure closing Cursor on table {0}",
                cause, i_tableName
            );
        }
    }

    /**Returns the SQL query string used for this Cursor (for diagnostic purposes).
      For normal queries this is only the WHERE-clause,for queries by example this is the SQL command pattern with ? as placeholders
      for parameters and a list of actual parameters in [...]. Note that for
      polymorphic queries the result is the SQL string lastly used.
    */ //2001-06-15 Knabe
    public String getQuery(){
        if(qbeObject!=null){ //use DiagnosticPreparedStatement:
            return qbeStmt.getSql() + " " + qbeStmt.getParams();
        }else{ //use conventional Statement:
            return query;
        }
    }

    private Table     table;
    private Session   session;
    private int       nTables;
    private ResultSet resultSet;
    private String    query;
    private Statement stmt;
    private Object    currObject;
    private Object    qbeObject;
    private DiagnosticPreparedStatement qbeStmt;
}

