/*
 * DiagnosticPreparedStatement.java
 *
 * Created on 8. Mrz 2001, 10:45
 */

package jora;

/**
 * Wraps a java.sql.PreparedStatement and information worth to be output in
 * the case of a failure, e.g. the SQL command and the parameters for it.
 * @author  Christoph Knabe 2001-05-16
 * @version 2002-06-24 Added checks for SQLWarning
 */
public class DiagnosticPreparedStatement {

    private static final String _where = DiagnosticPreparedStatement.class.getName();
    private final java.sql.PreparedStatement _stmt;
    private final String _sql;
    private final java.util.List _params = new java.util.ArrayList();

    /** Creates a new DiagnosticPreparedStatement with the given sql string
      * on the given connection.
    */
    public DiagnosticPreparedStatement(
        final java.sql.Connection i_connection,
        final String i_sql
    ) throws java.sql.SQLException
    {
        _stmt = i_connection.prepareStatement(i_sql);
        _sql = i_sql;
        SqlWarningFailure.check(i_connection);
    }

    /**Returns the SQL string pattern used by this PreparedStatement*/
    public String getSql(){return _sql;}

    /**Returns the parameters lastly given to this PreparedStatement*/
    public java.util.List getParams(){
        return java.util.Collections.unmodifiableList(_params);
    }

    public String toString(){
        return getSql() + "\n" + getParams();
    }

    /*package*/ int executeUpdate()throws java.sql.SQLException{
        final int result = _stmt.executeUpdate();
        SqlWarningFailure.check(_stmt);
        Failure.log(_where, this.toString());
        return result;
    }

    /*package*/ void clearParameters()throws java.sql.SQLException{
        _params.clear();
        _stmt.clearParameters();
        SqlWarningFailure.check(_stmt);
    }

    /*package*/ void addBatch()throws java.sql.SQLException{
        _stmt.addBatch();
        SqlWarningFailure.check(_stmt);
    }

    /*package*/ int[] executeBatch()throws java.sql.SQLException{
        final int[] result = _stmt.executeBatch();
        SqlWarningFailure.check(_stmt);
        return result;
    }

    /*package*/ java.sql.ResultSet executeQuery()throws java.sql.SQLException{
        final java.sql.ResultSet result = _stmt.executeQuery();
        SqlWarningFailure.check(_stmt);
        Failure.log(_where, this.toString());
        return result;
    }

    /*package*/ void setNull(final int i_index, final int i_sqlType)
    throws java.sql.SQLException
    {
        _setParam(i_index, null);
        _stmt.setNull(i_index, i_sqlType);
        SqlWarningFailure.check(_stmt);
    }

    /*package*/ void setByte(final int i_index, final byte i_param)
    throws java.sql.SQLException
    {
        _setParam(i_index, new Byte(i_param));
        _stmt.setByte(i_index, i_param);
        SqlWarningFailure.check(_stmt);
    }
    /*package*/ void setShort(final int i_index, final short i_param)
    throws java.sql.SQLException
    {
        _setParam(i_index, new Short(i_param));
        _stmt.setShort(i_index, i_param);
        SqlWarningFailure.check(_stmt);
    }
    /*package*/ void setInt(final int i_index, final int i_param)
    throws java.sql.SQLException
    {
        _setParam(i_index, new Integer(i_param));
        _stmt.setInt(i_index, i_param);
        SqlWarningFailure.check(_stmt);
    }
    /*package*/ void setLong(final int i_index, final long i_param)
    throws java.sql.SQLException
    {
        _setParam(i_index, new Long(i_param));
        _stmt.setLong(i_index, i_param);
        SqlWarningFailure.check(_stmt);
    }
    /*package*/ void setFloat(final int i_index, final float i_param)
    throws java.sql.SQLException
    {
        _setParam(i_index, new Float(i_param));
        _stmt.setFloat(i_index, i_param);
        SqlWarningFailure.check(_stmt);
    }
    /*package*/ void setDouble(final int i_index, final double i_param)
    throws java.sql.SQLException
    {
        _setParam(i_index, new Double(i_param));
        _stmt.setDouble(i_index, i_param);
        SqlWarningFailure.check(_stmt);
    }
    /*package*/ void setBoolean(final int i_index, final boolean i_param)
    throws java.sql.SQLException
    {
        _setParam(i_index, new Boolean(i_param));
        _stmt.setBoolean(i_index, i_param);
        SqlWarningFailure.check(_stmt);
    }
    /*package*/ void setBigDecimal(final int i_index, final java.math.BigDecimal i_param)
    throws java.sql.SQLException
    {
        _setParam(i_index, i_param);
        _stmt.setBigDecimal(i_index, i_param);
        SqlWarningFailure.check(_stmt);
    }
    /*package*/ void setString(final int i_index, final String i_param)
    throws java.sql.SQLException
    {
        _setParam(i_index, i_param);
        _stmt.setString(i_index, i_param);
        SqlWarningFailure.check(_stmt);
    }
    /*package*/ void setBytes(final int i_index, final byte[] i_param)
    throws java.sql.SQLException
    {
        _setParam(i_index, i_param);
        _stmt.setBytes(i_index, i_param);
        SqlWarningFailure.check(_stmt);
    }
    /*package*/ void setDate(final int i_index, final java.sql.Date i_param)
    throws java.sql.SQLException
    {
        _setParam(i_index, i_param);
        _stmt.setDate(i_index, i_param);
        SqlWarningFailure.check(_stmt);
    }
    /*package*/ void setTime(final int i_index, final java.sql.Time i_param)
    throws java.sql.SQLException
    {
        _setParam(i_index, i_param);
        _stmt.setTime(i_index, i_param);
        SqlWarningFailure.check(_stmt);
    }
    /*package*/ void setTimestamp(final int i_index, final java.sql.Timestamp i_param)
    throws java.sql.SQLException
    {
        _setParam(i_index, i_param);
        _stmt.setTimestamp(i_index, i_param);
        SqlWarningFailure.check(_stmt);
    }
    /*package*/ void setBinaryStream(final int i_index, final java.io.InputStream i_param, final int i_length)
    throws java.sql.SQLException
    {
        _setParam(i_index, i_param);
        _stmt.setBinaryStream(i_index, i_param, i_length);
        SqlWarningFailure.check(_stmt);
    }
    /*package*/ void setBlob(final int i_index, final java.sql.Blob i_param)
    throws java.sql.SQLException
    {
        _setParam(i_index, i_param);
        _stmt.setBlob(i_index, i_param);
        SqlWarningFailure.check(_stmt);
    }
    /*package*/ void setClob(final int i_index, final java.sql.Clob i_param)
    throws java.sql.SQLException
    {
        _setParam(i_index, i_param);
        _stmt.setClob(i_index, i_param);
        SqlWarningFailure.check(_stmt);
    }
    /*package*/ void close()
    throws java.sql.SQLException
    {
        _stmt.close();
        _params.clear();
    }

    private void _setParam(final int i_index, final Object i_value){
        _params.add(new _Param(i_index, i_value));
    }

    private class _Param {
        private final int _index;
        private final Object _value;

        /*package*/ _Param(final int i_index, final Object i_value){
            _index = i_index;   _value = i_value;
        }

        public String toString(){return "" +_index + "->" + _value;}
    }

}