//-< Session.java >--------------------------------------------------*--------*
// JORA                       Version 2.0        (c) 1998  GARRET    *     ?  *
// (Java Object Relational Adapter)                                  *   /\|  *
//                                                                   *  /  \  *
//                          Created:     10-Jun-98    K.A. Knizhnik  * / [] \ *
//                          Last update: 20-Jun-98    K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Database session abstraction
//-------------------------------------------------------------------*--------*

package jora;

import java.util.*;
import java.sql.*;

/**
 * This class is reponsible for establishing connection with database
 * and handling database errors.
 */
public class Session {
    public Connection connection;

    protected String    driver;   // driver class name
    protected Hashtable preparedStmtHash;
    protected int       connectionID;

    /** Session constructor
     *
     * @param driverClassName name of the database driver class
     */
    public Session(final String driverClassName)
    {
        driver = driverClassName;
        preparedStmtHash = new Hashtable();
	    connectionID = 0;
    }

    /** Session constructor
     *
     * @param driverClass class of database driver, must implement java.sql.Driver
     * @author Christoph Knabe 2003-03-12
     */
    public Session(final Class driverClass) { this(driverClass.getName()); }

    /** Session consructor for ODBC bridge driver
     */
    public Session() { this(sun.jdbc.odbc.JdbcOdbcDriver.class); }

    /** Default handler of database session errors.
     *
     * @param ex exception thrown by some of JDBC methods
     */
    public static void handleSQLException(SQLException ex) throws JdbcFailure {
        // A SQLException was generated.  Catch it and
        // rethrow ist encapsulating its error information
        // into a jora.Failure.
        //Note: The possible chain of causing exceptions will be
        //reported correctly by the framework MulTEx.
	      throw new JdbcFailure(ex);
    }

    /** Open database session.
     *  Attempt to establish a connection to the given database URL.
     *  The DriverManager attempts to select an appropriate driver from
     *  the set of registered JDBC drivers.
     *
     * @param url a database url of the form
     *  jdbc:<em>subprotocol</em>:<em>subname</em>
     * @param user the database user on whose behalf the Connection is being made
     * @param password the user's password
     * @throws OpenFailure if session is not succesfully opened
     */
    public void open(String dataSource, String user, String password)
    throws OpenFailure
    {
        try {
            Class.forName(driver);

            connection =
              DriverManager.getConnection(dataSource, user, password);
            connectionID += 1;
       }
       catch(SQLException ex) {
            throw new OpenFailure(new JdbcFailure(ex), dataSource, user, driver);
       }
       catch(Exception ex) {
            throw new OpenFailure(ex, dataSource, user, driver);
       }
    }

    /** Close only the PreparedStatement-s of the database session.
      * With ACCESS necessary in order to drop the tables. Knabe 2001-03-09
    */
    public void closePreparedStatements() throws java.sql.SQLException {
        final Enumeration items = preparedStmtHash.elements();
        while (items.hasMoreElements()) {
            final DiagnosticPreparedStatement pstmt
            = (DiagnosticPreparedStatement)items.nextElement();
            pstmt.close();
        }
        preparedStmtHash.clear();
        preparedStmtHash = null;
    }

    /** Close database session and release all resources held by session.
     */
    public void close() throws CloseFailure {
        try {
            closePreparedStatements();
            connection.close();
            connection = null;
            driver = null;
        }
        catch (SQLException ex) { throw new CloseFailure(new JdbcFailure(ex)); }
    }

    /**
     * Execute a SQL INSERT, UPDATE or DELETE statement. In addition,
     * SQL statements that return nothing such as SQL DDL statements
     * can be executed.
     *
     * @param sql a SQL INSERT, UPDATE or DELETE statement or a SQL
     * statement that returns nothing
     * @return either the row count for INSERT, UPDATE or DELETE or 0
     * for SQL statements that return nothing
     */
    public int execute(String sql) {
        try {
            Statement stmt = connection.createStatement();
            SqlWarningFailure.check(connection);
            Failure.log("executeUpdate", sql); //Knabe 01-02-15
            int result = stmt.executeUpdate(sql);
            SqlWarningFailure.check(stmt);
            stmt.close();
            return result;
        } catch(SQLException ex) { throw new SqlFailure(new JdbcFailure(ex), sql); } //Knabe 01-02-27
        //return -1;
    }

    /**
     * Commit makes all changes made since the previous
     * commit/rollback permanent and releases any database locks
     * currently held by the Connection. This method should only be
     * used when auto commit has been disabled.
     */
    public void commit() {
        try {
            connection.commit();
        } catch(SQLException ex) { handleSQLException(ex); }
        SqlWarningFailure.check(connection);
    }

    /**
     * Rollback drops all changes made since the previous
     * commit/rollback and releases any database locks currently held
     * by the Connection. This method should only be used when auto
     * commit has been disabled.
     */
    public void rollback() {
        try {
            connection.rollback();
        } catch(SQLException ex) { handleSQLException(ex); }
        SqlWarningFailure.check(connection);
    }

    //2001-04-19  Christoph Knabe  With unnamed exception parameters
    //2001-02-26  Christoph Knabe  Creation
    /** A JORA operation failed because a SQL statement execution failed.
    */
    public static class OpenFailure extends Failure {
        public OpenFailure(
            final Exception i_cause,
            final String i_url, final String i_user, final String i_driver
        ){
            super("Unable to open database session with URL {0}, user \"{1}\" and driver {2}",
              i_cause, i_url, i_user, i_driver
            );
        }
    }

    //2001-04-19  Christoph Knabe  With unnamed exception parameters
    //2001-02-26  Christoph Knabe  Creation
    /** A JORA operation failed because a SQL statement execution failed.
    */
    public static class CloseFailure extends Failure {
        public CloseFailure(final Exception i_cause){
            super("Unable to correctly close the database session ", i_cause);
        }
    }
}//Session
