//-< Table.java >----------------------------------------------------*--------*
// JORA                       Version 2.0        (c) 1998  GARRET    *     ?  *
// (Java Object Relational Adapter)                                  *   /\|  *
//                                                                   *  /  \  *
//                          Created:     10-Jun-98    K.A. Knizhnik  * / [] \ *
//                          Last update: 20-Jun-98    K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Class representing database table
//-------------------------------------------------------------------*--------*

//2001-07-06  Knabe  Made robust against forget of creating a database Session
//                   in the Table init() method.

package jora;

import java.util.*;
import java.sql.*;
import java.lang.reflect.*;

/** Table class is used to establish mapping between corteges of database
 *  tables and Java classes. This class is responsible for constructing
 *  SQL statements for extracting, updating and deleting records of the
 *  database table.
 * @version 2002-06-24 Added checks for SQLWarning
 */
public class Table {
    /** Constructor for table object. Make association between Java class
     *  and database table.
     *
     * @param className name of Java class
     * @param tableName name of database table mapped on this Java class
     * @param key table's primary key. This parameter is used in UPDATE/DELETE
     *  operations to locate record in the table.
     * @s session, which should be opened before first access to the table
     */
    public Table(String className, String tableName, Session s, String key) {
	String[] keys = {key};
	init(className, tableName, s, keys);
    }

    /** Constructor for table object. Make association between Java class
     *  and database table.
     *
     * @param className name of Java class
     * @param tableName name of database table mapped on this Java class
     * @param keys table primary keys. This parameter is used in UPDATE/DELETE
     *  operations to locate record in the table.
     * @s session, which should be opened before first access to the table
     */
    public Table(String className, String tableName, Session s, String[] keys)
    {
	init(className, tableName, s, keys);
    }

    /** Constructor for table object. Make association between Java class
     *  and database table. Name of Java class should be the same as name of
     *  the database table
     *
     * @param className name of Java class, which should be (without
     *  package prefix) be the same as the name of database table.
     * @param keys table primary keys. This parameter is used in UPDATE/DELETE
     *  operations to locate record in the table.
     * @s session, which should be opened before first access to the table
     */
    public Table(String className, Session s, String[] keys) {
        init(className, className.substring(className.lastIndexOf('.')+1),
	     s, keys);
    }

    /** Constructor for table object. Make association between Java class
     *  and database table. Name of Java class should be the same as name of
     *  the database table
     *
     * @param className name of Java class, which should be (without
     *  package prefix) be the same as the name of database table.
     * @param key table primary key. This parameter is used in UPDATE/DELETE
     *  operations to locate record in the table.
     * @s session, which should be opened before first access to the table
     */
    public Table(String className, Session s, String key) {
	String[] keys = {key};
        init(className, className.substring(className.lastIndexOf('.')+1),
	     s, keys);
    }

    /** Constructor of table without explicit key specification.
     *  Specification of key is necessary for update/remove operations.
     *  If key is not specified, it is inherited from base table (if any).
     */
    public Table(String className, Session s) {
        init(className, className.substring(className.lastIndexOf('.')+1),
	     s, null);
    }


    /** Constructor of table with "key" and "session" parameters inherited
     *  from base table.
     */
    public Table(String className) {
        init(className, className.substring(className.lastIndexOf('.')+1),
	     null, null);
    }

    /**Get the name of the table*/
    public String name(){return name;}

    /** Select records from database table according to search condition
     *
     * @param condition valid SQL condition expression started with WHERE
     *  or empty string if all records should be fetched.
     */
    public final Cursor select(String condition) {
        return new Cursor(this, session, 1, condition);
    }

    /** Select records from database table according to search condition
     *
     * @param condition valid SQL condition expression started with WHERE
     *  or empty string if all records should be fetched.
     * @param session user database session
     */
    public final Cursor select(String condition, Session session) {
        return new Cursor(this, session, 1, condition);
    }

    /** Select records from specified and derived database tables
     *
     * @param condition valid SQL condition expression started with WHERE
     *  or empty string if all records should be fetched.
     */
    public final Cursor selectAll(String condition) {
        return new Cursor(this, session, nDerived+1, condition);
    }

    /** Select records from specified and derived database tables
     *
     * @param condition valid SQL condition expression started with WHERE
     *  or empty string if all records should be fetched.
     * @param session user database session
     */
    public final Cursor selectAll(String condition, Session session) {
        return new Cursor(this, session, nDerived+1, condition);
    }


    /** Select records from database table using <I>obj</I> object as
     * template for selection. All non-builtin fields of this object,
     * which are not null, are compared with correspondent table values.
     *
     * @param obj object for construction search condition: selected objects
     *  should match all non-null fields of specified object.
     */
    public final Cursor queryByExample(Object obj) {
        return new Cursor(this, session, 1, obj);
    }

    /** Select records from database table using <I>obj</I> object as
     * template for selection. All non-builtin fields of this object,
     * which are not null, are compared with correspondent table values.
     *
     * @param obj object for construction search condition: selected objects
     *  should match all non-null fields of specified object.
     * @param session user database session
     */
    public final Cursor queryByExample(Object obj, Session session) {
        return new Cursor(this, session, 1, obj);
    }

    /** Select records from specified and derived database tables using
     * <I>obj</I> object as template for selection.
     * All non-builtin fields of this object,
     * which are not null, are compared with correspondent table values.
     *
     * @param obj object for construction search condition: selected objects
     *  should match all non-null fields of specified object.
     */
    public final Cursor queryAllByExample(Object obj) {
        return new Cursor(this, session, nDerived+1, obj);
    }

    /** Select records from specified and derived database tables using
     * <I>obj</I> object as template for selection.
     * All non-builtin fields of this object,
     * which are not null, are compared with correspondent table values.
     *
     * @param obj object for construction search condition: selected objects
     *  should match all non-null fields of specified object.
     * @param session user database session
     */
    public final Cursor queryAllByExample(Object obj, Session session) {
        return new Cursor(this, session, nDerived+1, obj);
    }

    /** Insert new record in the table. Values of inserted record fields
     *  are taken from the specified object.
     *
     * @param obj object specifing values of inserted record fields
     */
    public void insert(Object obj) throws InsertFailure {
	insert(obj, session);
    }

    /** Insert several new records in the table. Values of inserted records
     *  fields are taken from objects of specified array.
     *
     * @param objects array with objects specifing values of inserted record
     * fields
     */
    public void insert(Object[] objects) throws InsertFailure {
        insert(objects, session);
    }

    /** Insert new record in the table using specified database session.
     *  Values of inserted record fields
     *  are taken from the specified object.
     *
     * @param obj object specifing values of inserted record fields
     * @param session user database session
     */
    public synchronized void insert(Object obj, Session session)
    throws InsertFailure
    {
        try {
                _prepareInsertStmt(session);
           bindUpdateVariables(insertStmt, obj);
                Failure.log("executeUpdate", "insert ..."); //Knabe 01-02-15
           insertStmt.executeUpdate();
           insertStmt.clearParameters();
       }
       catch(SQLException ex){throw new InsertFailure(new JdbcFailure(ex), insertStmt);}
       catch(Exception ex){throw new InsertFailure(ex, insertStmt);}
    }

    /** Insert several new records in the table. Values of inserted records
     *  fields are taken from objects of specified array.
     *
     * @param objects array with objects specifing values of inserted record
     *                fields
     * @param session user database session
     */
    public synchronized void insert(Object[] objects, Session session)
    throws InsertFailure
    //OUT insertStmt
    {
        try {
            _prepareInsertStmt(session);
           for (int i = 0; i < objects.length; i++) {
               bindUpdateVariables(insertStmt, objects[i]);
               insertStmt.addBatch();
           }
            Failure.log("executeBatch", "insert ..."); //Knabe 01-02-15
           insertStmt.executeBatch();
           insertStmt.clearParameters();
       }
       catch(SQLException ex) { throw new InsertFailure(new JdbcFailure(ex), insertStmt); }
       catch(Exception ex) { throw new InsertFailure(ex, insertStmt); }
    }

    /** Sets up the PreparedStatement for inserting.
      * The global variable insertStmt gets a reference to it.
    */
    private void _prepareInsertStmt(Session session) throws SQLException
    //OUT insertStmt
    {
        if (session == null) {
            session = ((SessionThread)Thread.currentThread()).session;
        }
        checkConnection(session);
        if (insertStmt == null) {
            final StringBuffer buf = new StringBuffer("insert into ");
            buf.append(name);   buf.append(" (");   buf.append(listOfFields);
            buf.append(") values (?");
            for (int i = 1; i < nColumns; i++) {
                 buf.append(",?");
            }
            buf.append(")");
            final String sql = buf.toString();
            Failure.log("prepareStatement", sql); //Knabe 01-02-15
            insertStmt = new DiagnosticPreparedStatement(session.connection, sql);
        }
    }

    /** Update record in the table using table's primary key to locate
     *  record in the table and values of fields of specified object <I>obj</I>
     *  to alter record fields.
     *
     * @param obj object specifing value of primary key and new values of
     *  updated record fields
     *
     * @return number of objects actually updated
    */
    public int update(Object obj)
    throws NoPrimaryKeyFailure, UpdateFailure
    {
	return update(obj, session);
    }

    /** Update record in the table using table's primary key to locate
     *  record in the table and values of fields of specified object <I>obj</I>
     *  to alter record fields.
     *
     * @param obj object specifing value of primary key and new values of
     *  updated record fields
     * @param session user database session
     *
     * @return number of objects actually updated
     */
    public synchronized int update(Object obj, Session session)
    throws NoPrimaryKeyFailure, UpdateFailure
    {
      	int nUpdated = 0;
        try {
            _prepareUpdateStmt(session);

           for (int i = 0; i < primaryKeys.length; i++) {
          fields[primaryKeyIndices[i]].bindVariable(updateStmt, obj,
							  nColumns+i+1);
           }
           bindUpdateVariables(updateStmt, obj);
           nUpdated = updateStmt.executeUpdate();
           updateStmt.clearParameters();
       }
       catch(SQLException ex) {throw new UpdateFailure(new JdbcFailure(ex), updateStmt);}
       catch(Exception ex) {throw new UpdateFailure(ex, updateStmt);}
       return nUpdated;
    }

    /** Update set of records in the table using table's primary key to locate
     *  record in the table and values of fields of objects from sepecifed
     *  array <I>objects</I>  to alter record fields.
     *
     * @param objects array of objects specifing primary keys and and new
     * values of updated record fields
     *
     * @return number of objects actually updated
     */
    public int update(Object[] objects) {
	return update(objects, session);
    }

    /** Update set of records in the table using table's primary key to locate
     *  record in the table and values of fields of objects from sepecifed
     *  array <I>objects</I>  to alter record fields.
     *
     * @param objects array of objects specifing primiray keys and and new
     * values of updated record fields
     * @param session user database session
     *
     * @return number of objects actually updated
     */
    public synchronized int update(Object[] objects, Session session) {
        int nUpdated = 0;
        try {
                _prepareUpdateStmt(session);

           for (int i = 0; i < objects.length; i++) {
          for (int j = 0; j < primaryKeys.length; j++) {
              fields[primaryKeyIndices[j]].bindVariable(updateStmt,
                               objects[i],
                               nColumns+1+j);
          }
               bindUpdateVariables(updateStmt, objects[i]);
               updateStmt.addBatch();
           }
           int rc[] = updateStmt.executeBatch();
           for (int k = 0; k < rc.length; k++) {
          nUpdated += rc[k];
           }
           updateStmt.clearParameters();
       }
       catch(SQLException ex) {throw new UpdateFailure(new JdbcFailure(ex), updateStmt);}
       catch(Exception ex) {throw new UpdateFailure(ex, updateStmt);}
       return nUpdated;
    }

    /** Sets up the PreparedStatement for updating.
      * The global variable updateStmt gets a reference to it.
    */
    private void _prepareUpdateStmt(Session session) throws SQLException
    //OUT updateStmt
    {
        if (primaryKeys == null) {
            throw new NoPrimaryKeyFailure(null);
        }
        if (session == null) {
            session = ((SessionThread)Thread.currentThread()).session;
        }
        checkConnection(session);
        if (updateStmt == null) {
            final StringBuffer buf = new StringBuffer("update ");
            buf.append(name);   buf.append(" set ");
            buf.append(listOfAssignments);   buf.append(" where ");
            for (int i = 0; i < primaryKeys.length; i++) {
                if(i!=0){buf.append(" and ");}
                buf.append(primaryKeys[i]);   buf.append(" = ?");
            }
            final String sql = buf.toString();
            updateStmt = new DiagnosticPreparedStatement(session.connection, sql);
        }
    }


    /** Delete record with specified value of primary key from the table.
     *
     * @param obj object containing value of primary key.
     *
     * @return number of objects actually deleted
     */
    public int delete(Object obj) {
	return delete(obj, session);
    }

    /** Delete record with specified value of primary key from the table.
     *
     * @param obj object containing value of primary key.
     * @param session user database session
     */
    public synchronized int delete(Object obj, Session session)
    throws NoPrimaryKeyFailure, DeleteFailure
    {
        try {
            _prepareDeleteStmt(session);
	    for (int i = 0; i < primaryKeys.length; i++) {
		fields[primaryKeyIndices[i]].bindVariable(deleteStmt, obj,i+1);
	    }
	    final int nDeleted = deleteStmt.executeUpdate();
	    deleteStmt.clearParameters();
            return nDeleted;
       }
       catch(SQLException ex) {throw new DeleteFailure(new JdbcFailure(ex), deleteStmt);}
       catch(Exception ex) {throw new DeleteFailure(ex, deleteStmt);}
    }


    /** Delete records with specified primary keys from the table.
     *
     * @param objects array of objects containing values of primary key.
     *
     * @return number of objects actually deleted
     */
    public int delete(Object[] objects) {
	return delete(objects, session);
    }

    /** Delete records with specified primary keys from the table.
     *
     * @param objects array of objects containing values of primary key.
     *
     * @return number of objects actually deleted
     */
    public synchronized int delete(Object[] objects, Session session)
    throws NoPrimaryKeyFailure
    {
        try {
            _prepareDeleteStmt(session);
	    for (int i = 0; i < objects.length; i++) {
		for (int j = 0; j < primaryKeys.length; j++) {
		    fields[primaryKeyIndices[j]].bindVariable(deleteStmt,
							      objects[i], j+1);
		}
		deleteStmt.addBatch();
	    }
	    final int rc[] = deleteStmt.executeBatch();
            int nDeleted = 0;
	    for (int k = 0; k < rc.length; k++) {
		nDeleted += rc[k];
	    }
	    deleteStmt.clearParameters();
            return nDeleted;
       }
       catch(SQLException ex) {throw new DeleteFailure(new JdbcFailure(ex), deleteStmt);}
       catch(Exception ex) {throw new DeleteFailure(ex, deleteStmt);}
    }

    /** Sets up the PreparedStatement for deleting.
      * The global variable deleteStmt gets a reference to it.
    */
    private void _prepareDeleteStmt(Session session) throws SQLException
    {
        if (primaryKeys == null) {
            throw new NoPrimaryKeyFailure(null);
        }
        if (session == null) {
            session = ((SessionThread)Thread.currentThread()).session;
        }
        checkConnection(session);
        if (deleteStmt == null) {
            final StringBuffer buf = new StringBuffer("delete from ");
            buf.append(name);   buf.append(" where ");
            for (int i = 0; i < primaryKeys.length; i++) {
                if(i!=0){buf.append(" and ");}
                buf.append(primaryKeys[i]);   buf.append(" = ?");
            }
            final String sql = buf.toString();
            deleteStmt = new DiagnosticPreparedStatement(session.connection, sql);
        }
    }

    /** Spearator of name components of compound field. For example, if Java
     *  class constains component "location" of Point class, which
     *  has two components "x" and "y", then database table should
     *  have columns "location_x" and "location_y" (if '_' is used
     *  as separator)
     */
    public static String fieldSeparator = "_";


    /** Some versions of JDBC driver doesn't support
     *  <code>getBigDecimal(int columnIndex)</code>. Setting this variable to
     *  <code>true</code> makes JORA to explicitly request scale from result
     *  set metadata and use deprecated version of <code>getBigDecimal</code>
     *  with extra <code>scale</code> parameter.
     */
    public static boolean useDepricatedGetBigDecimal = true;


    //2001-04-19  Christoph Knabe  With unnamed exception parameters
    //2001-02-26  Christoph Knabe  Creation
    /** An insert operation failed because a SQL statement execution failed.
    */
    public class InsertFailure extends Failure {
        public InsertFailure(
            final Exception i_cause,
            final jora.DiagnosticPreparedStatement i_pstmt
            //final String i_sql,
            //final String i_params
        ){
            super("Unable to insert table row from object. "
                + "Details: 0=sql, 1=sqlParams",
                i_cause,
                (i_pstmt==null ? "" : i_pstmt.getSql()),
                (i_pstmt==null ? "" : i_pstmt.getParams().toString())
            );
            //???02-06-24:
            Failure.log("jora.Table-exception created", multex.Msg.getReport(this));
        }
    }

    //2001-04-19  Christoph Knabe  With unnamed exception parameters
    //2001-03-08  Christoph Knabe  Creation
    /** An update operation failed because a SQL statement execution failed.
    */
    public class UpdateFailure extends Failure {
        public UpdateFailure(final Exception i_cause,
        final jora.DiagnosticPreparedStatement i_pstmt
        ){
            super(
                "Unable to update table row from object. "
                + "Details: 0=sql, 1=sqlParams",
                i_cause, i_pstmt.getSql(), i_pstmt.getParams()
            );
        }
    }

    //2001-04-19  Christoph Knabe  With unnamed exception parameters
    //2001-03-08  Christoph Knabe  Creation
    /** A delete operation failed because a SQL statement execution failed.
    */
    public class DeleteFailure extends Failure {
        public DeleteFailure(
            final Exception i_cause,
            final jora.DiagnosticPreparedStatement i_pstmt
        ){
            super(
                "Unable to delete table row for object. "
                + "Details: 0=sql, 1=sqlParams",
                i_cause, i_pstmt.getSql(), i_pstmt.getParams()
            );
        }
    }

    // --- Implementation -----------------------------------------

    /** Is table abstract - not present in database.
     */
    protected boolean isAbstract;
    protected Table   derived;
    protected int     nDerived;

    /*protected*/private String  name; //of the table
    protected String  listOfFields;
    protected String  listOfAssignments;
    protected Class   cls;
    protected Session session;

    static private Class serializableClass;
    private   FieldDescriptor[] fields;

    private   int     nFields;  // length of "fields" array
    private   int     nColumns; // number of atomic fields in "fields" array

    private   String  primaryKeys[];
    private   int     primaryKeyIndices[];

    protected int     connectionID;

    private DiagnosticPreparedStatement updateStmt;
    private DiagnosticPreparedStatement deleteStmt;
    private DiagnosticPreparedStatement insertStmt;

    private static Table  allTables;
    private Constructor   constructor;
    private static Method setBypass;

    private static final Object[] bypassFlag = { new Boolean(true) };
    private static final Object[] constructorArgs = {};

    static {
        try {
	    serializableClass = java.io.Serializable.class; //Knabe 02-06-22
       final Class c = java.lang.reflect.AccessibleObject.class; //Knabe 02-06-22
	    final Class[] param = { Boolean.TYPE };
	    setBypass = c.getMethod("setAccessible", param);
        } catch(Exception ex) {throw new StartupFailure(ex); /*Knabe 01-02-27*/}
    }


    private final void init(String className, String tableName, Session s,
			    String[] keys)
    throws InitFailure
    {
        try {
            name = tableName;
            cls = Class.forName(className);
            isAbstract = tableName == null;
            session = s;
            primaryKeys = keys;
            listOfFields = "";
            listOfAssignments = "";
            connectionID = 0;
            Vector fieldsVector = new Vector();
            nFields = buildFieldsList(fieldsVector, cls, "");
            fields = new FieldDescriptor[nFields];
            fieldsVector.copyInto(fields);

            try {
                constructor = cls.getDeclaredConstructor(new Class[0]);
                setBypass.invoke(constructor, bypassFlag);
            } catch(Exception ex) {throw new Failure( /*Knabe 01-02-20*/
                ex, "Unable to locate default constructor for class", className
            );}

            if (keys != null) {
                if (keys.length == 0) {
                    throw new NoPrimaryKeyFailure(null);
                }
                primaryKeyIndices = new int[keys.length];
                for (int j = keys.length; --j >= 0;) {
                    final String key = keys[j]; //Knabe 01-02-28
                    int i = nFields;
                    while (--i >= 0) {
                        if (fields[i].name.equals(key)) {
                            if (!fields[i].isAtomic()) {
                                throw new NoPrimaryKeyFailure(key);
                            }
                            primaryKeyIndices[j] = i;
                            break;
                        }
                    }
                    if (i < 0) {
                        throw new NoSuchFieldException("No such field '" + key
                                                   + "' for use as key in class " + className);
                    }
                }
            }
            insertIntoTableHierarchy();
        } catch(Exception ex) { //Knabe 2001-02-27
            final StringBuffer keysBuf = new StringBuffer();
            if(keys!=null){
                for(int i=0; i<keys.length; i++){
                    if(i>0){keysBuf.append(',');}
                    keysBuf.append(keys[i]);
                }
            }
            throw new InitFailure(//Knabe 01-02-27
                ex,  className, tableName, keysBuf.toString()
            );
        }
    }


    private final void insertIntoTableHierarchy()
    {
	Table t, prev = null;
	Table after = null;
	int nChilds = 0;
	for (t = allTables; t != null; prev = t, t = t.derived) {
	    if (t.cls.isAssignableFrom(cls)) {
 	        if (primaryKeys == null && t.primaryKeys != null) {
		    primaryKeys = t.primaryKeys;
		    primaryKeyIndices = t.primaryKeyIndices;
		}
		if (session == null) {
		    session = t.session;
		}
		t.nDerived += 1;
		after = t;
	    } else if (cls.isAssignableFrom(t.cls)) {
		after = prev;
		do {
		    if (cls.isAssignableFrom(t.cls)) {
		        if (primaryKeys != null && t.primaryKeys == null) {
			    t.primaryKeys = primaryKeys;
			    t.primaryKeyIndices = primaryKeyIndices;
			}
			if (t.session == null) {
			    t.session = session;
			}
			nChilds += 1;
		    }
		} while ((t = t.derived) != null);
		break;
	    }
	}
	if (after == null) {
	    derived = allTables;
	    allTables = this;
	} else {
	    derived = after.derived;
	    after.derived = this;
	}
	nDerived = nChilds;
    }

    private final void checkConnection(Session s) throws SQLException {
        //Knabe 2001-07-06 check inserted:
        if(s==null){throw new multex.Failure(  //knabe 2001-07-06
            "Missing Session object for the database connection", null
        );}
        //Knabe
        if (connectionID != s.connectionID) {
	    if (insertStmt != null) {
	        insertStmt.close();
	        insertStmt = null;
  	    }
	    if (updateStmt != null) {
	        updateStmt.close();
	        updateStmt = null;
	    }
	    if (deleteStmt != null) {
	        deleteStmt.close();
	        deleteStmt = null;
	    }
	    connectionID = s.connectionID;
        }
    }


    private final int buildFieldsList(Vector buf, Class cls, String prefix)
    {
	Field[] f = cls.getDeclaredFields();

	Class superclass = cls;
	while ((superclass = superclass.getSuperclass()) != null) {
	    Field[] inheritedFields = superclass.getDeclaredFields();
	    Field[] allFields = new Field[inheritedFields.length + f.length];
	    System.arraycopy(inheritedFields, 0, allFields, 0,
			     inheritedFields.length);
	    System.arraycopy(f,0, allFields, inheritedFields.length, f.length);
	    f = allFields;
	}

	try {
	    for (int i = f.length; --i>= 0;) {
	        setBypass.invoke(f[i], bypassFlag);
	    }
	} catch(Exception ex) {
	    //System.err.println("Failed to set bypass attribute");
            //Knabe 01-02-21:
            throw new Failure(ex, "Failed to set bypass attribute");
	}

	int n = 0;
	for (int i = 0; i < f.length; i++) {
	    if ((f[i].getModifiers()&(Modifier.TRANSIENT|Modifier.STATIC))==0)
	    {
		String name = f[i].getName();
		Class fieldClass = f[i].getType();
		String fullName = prefix + name;
		FieldDescriptor fd = new FieldDescriptor(f[i], fullName);
		int type;

		buf.addElement(fd);
		n += 1;

		String c = fieldClass.getName();
		if (c.equals("byte")) type = FieldDescriptor.t_byte;
		else if (c.equals("short")) type = FieldDescriptor.t_short;
		else if (c.equals("int")) type = FieldDescriptor.t_int;
		else if (c.equals("long")) type = FieldDescriptor.t_long;
		else if (c.equals("float")) type = FieldDescriptor.t_float;
		else if (c.equals("double")) type = FieldDescriptor.t_double;
		else if (c.equals("boolean")) type = FieldDescriptor.t_boolean;
		else if (c.equals("java.lang.Byte"))
		    type = FieldDescriptor.tByte;
		else if (c.equals("java.lang.Short"))
		    type = FieldDescriptor.tShort;
		else if (c.equals("java.lang.Integer"))
		    type = FieldDescriptor.tInteger;
		else if (c.equals("java.lang.Long"))
		    type = FieldDescriptor.tLong;
		else if (c.equals("java.lang.Float"))
		    type = FieldDescriptor.tFloat;
		else if (c.equals("java.lang.Double"))
		    type = FieldDescriptor.tDouble;
		else if (c.equals("java.lang.Boolean"))
		    type = FieldDescriptor.tBoolean;
		else if (c.equals("java.math.BigDecimal"))
		    type = FieldDescriptor.tDecimal;
		else if (c.equals("java.lang.String"))
		    type = FieldDescriptor.tString;
		else if (c.equals("java.lang.[B"))
  		    type = FieldDescriptor.tBytes;
		else if (c.equals("java.sql.Date"))
		    type = FieldDescriptor.tDate;
		else if (c.equals("java.sql.Time"))
		    type = FieldDescriptor.tTime;
		else if (c.equals("java.sql.Timestamp"))
		    type = FieldDescriptor.tTimestamp;
		else if (c.equals("java.lang.InputStream"))
		    type = FieldDescriptor.tStream;
		else if (c.equals("java.sql.BlobLocator"))
		    type = FieldDescriptor.tBlob;
		else if (c.equals("java.sql.ClobLocator"))
		    type = FieldDescriptor.tClob;
		else if (serializableClass.isAssignableFrom(fieldClass))
 		    type = FieldDescriptor.tClosure;
                else {
		    int nComponents = buildFieldsList(buf, fieldClass,
						      fd.name+fieldSeparator);
		    fd.inType = fd.outType =
		        FieldDescriptor.tCompound + nComponents;

		    try {
		        fd.constructor = //locate default constructor
			  fieldClass.getDeclaredConstructor(new Class[0]);
			setBypass.invoke(fd.constructor, bypassFlag);
                    } catch(Exception ex) {throw new Failure( /*Knabe 01-02-21*/
                        ex, "Unable to locate default constructor"
                    );}

		    n += nComponents;
		    continue;
		}
		if (listOfFields.length() != 0) {
		    listOfFields += ",";
		    listOfAssignments += ",";
		}
		listOfFields += fullName;
		listOfAssignments += fullName + "=?";

		fd.inType = fd.outType = type;
		nColumns += 1;
	    }
	}
	return n;
    }


    protected final Object load(ResultSet resultSet) throws SQLException {
	Object obj;
        try {
	    obj = constructor.newInstance(constructorArgs);
	}
        //Knabe 2001-02-21 All Exception-s handled the same:
	//catch(IllegalAccessException ex) { throw new IllegalAccessError(); }
	//catch(InstantiationException ex) { throw new InstantiationError(); }
	catch(Exception ex) {
           //throw new InstantiationError("Exception was thrown by constructor");
           throw new Failure(ex, "Exception was thrown by constructor");
	}
	load(obj, 0, nFields, 0, resultSet);
	return obj;
    }

    private final int load(Object obj, int i, int end, int column,
			    ResultSet resultSet)
	throws SQLException
    {
	try {
	    while (i < end) {
	        FieldDescriptor fd = fields[i++];
		if (!fd.loadVariable(resultSet, obj, ++column)) {
		    Object component =
		        fd.constructor.newInstance(constructorArgs);
		    fd.field.set(obj, component);
		    int nComponents = fd.inType - FieldDescriptor.tCompound;
		    column = load(component, i, i + nComponents,
				  column-1, resultSet);
		    i += nComponents;
		}
	    }
	}
        //2001-02-21  Knabe  All Exception-s handled the same:
	//catch(IllegalAccessException ex) { throw new IllegalAccessError(); }
	//catch(InstantiationException ex) { throw new InstantiationError(); }
        //catch(InvocationTargetException ex) {
        catch(Exception ex) {
           //throw new InstantiationError("Exception was thrown by constructor");
           throw new Failure(ex, "Exception was thrown by constructor");
	}
	return column;
    }

    protected final void bindUpdateVariables(DiagnosticPreparedStatement pstmt,
					     Object                      obj)
      throws SQLException, java.io.IOException, IllegalAccessException
    {
        bindUpdateVariables(pstmt, obj, 0, nFields, 0);
    }

    protected final void bindQueryVariables(DiagnosticPreparedStatement pstmt,
					    Object            obj)
       throws SQLException, java.io.IOException, IllegalAccessException
    {
	bindQueryVariables(pstmt, obj, 0, nFields, 0);
    }


    protected final void updateVariables(ResultSet resultSet, Object obj)
       throws SQLException, IllegalAccessException, java.io.IOException
    {
        updateVariables(resultSet, obj, 0, nFields, 0);
        resultSet.updateRow();
        SqlWarningFailure.check(resultSet);
    }


    protected final String buildQueryList(Object qbe)
    {
        StringBuffer buf = new StringBuffer();
        buildQueryList(buf, qbe, 0, nFields);
        String condition = buf.toString();
        return condition.length() != 0 ? " where " + condition : "";
    }


    private final int bindUpdateVariables(DiagnosticPreparedStatement pstmt, Object obj,
					  int i, int end, int column)
        throws SQLException, java.io.IOException, IllegalAccessException
    {
        while (i < end) {
            FieldDescriptor fd = fields[i++];
            Object comp = null;
            if (!fd.isBuiltin() && (comp = fd.field.get(obj)) == null) {
                if (fd.isCompound()) {
                    int nComponents = fd.outType-FieldDescriptor.tCompound;
                    while (--nComponents >= 0) {
                        fd = fields[i++];
                        if (!fd.isCompound()) {
                            pstmt.setNull(++column,
                               FieldDescriptor.sqlTypeMapping[fd.outType]);
                        }
                    }
                } else {
                    pstmt.setNull(++column,
                                  FieldDescriptor.sqlTypeMapping[fd.outType]);
                }
            } else {
                if (!fd.bindVariable(pstmt, obj, ++column)) {
                    int nComponents = fd.outType-FieldDescriptor.tCompound;
                    column = bindUpdateVariables(pstmt, comp,
                                                 i,i+nComponents,column-1);
                    i += nComponents;
                }
            }
        }
	return column;
    }

    private final int bindQueryVariables(DiagnosticPreparedStatement pstmt, Object obj,
					 int i, int end, int column)
        throws SQLException, java.io.IOException, IllegalAccessException
    {
        while (i < end) {
            Object comp;
            FieldDescriptor fd = fields[i++];
            if (!fd.field.getDeclaringClass().isInstance(obj)) {
                return column;
            }
            int nComponents =
                fd.isCompound() ? fd.outType-FieldDescriptor.tCompound : 0;
            if (!fd.isBuiltin()
                && fd.outType != FieldDescriptor.tClosure
                && (comp = fd.field.get(obj)) != null)
            {
                if (!fd.bindVariable(pstmt, obj, ++column)) {
                    column = bindQueryVariables(pstmt, comp,
                                                i,i+nComponents,column-1);
                }
            }
            i += nComponents;
        }
	return column;
    }

    private final void buildQueryList(StringBuffer buf, Object qbe,
				      int i, int end)
    {
	try {
	    while (i < end) {
		Object comp;
		FieldDescriptor fd = fields[i++];
		int nComponents =
		    fd.isCompound() ? fd.outType-FieldDescriptor.tCompound : 0;
		if (!fd.isBuiltin()
		    && fd.outType != FieldDescriptor.tClosure
		    && (comp = fd.field.get(qbe)) != null)
		{
		    if (nComponents != 0) {
			buildQueryList(buf, comp, i, i+nComponents);
		    } else {
			if (buf.length() != 0) {
			    buf.append(",");
			}
			buf.append(fd.name);
			buf.append("=?");
		    }
		}
		i += nComponents;
	    }
	} catch(IllegalAccessException ex) {
            //throw new IllegalAccessError(); //2001-02-21  Knabe
            throw new Failure(ex);
        }
    }

    protected final int updateVariables(ResultSet resultSet, Object obj,
					 int i, int end, int column)
    throws SQLException, IllegalAccessException, java.io.IOException
    {
        //try {
        while (i < end) {
            FieldDescriptor fd = fields[i++];
            Object comp = null;
            if (!fd.isBuiltin() && (comp = fd.field.get(obj)) == null) {
                if (fd.isCompound()) {
                    int nComponents = fd.outType-FieldDescriptor.tCompound;
                    while (--nComponents >= 0) {
                        fd = fields[i++];
                        if (!fd.isCompound()) {
                            resultSet.updateNull(++column);
                            SqlWarningFailure.check(resultSet);
                        }
                    }
                } else {
                    resultSet.updateNull(++column);
                    SqlWarningFailure.check(resultSet);
                }
            } else {
                if (!fd.updateVariable(resultSet, obj, ++column)) {
                    int nComponents = fd.outType-FieldDescriptor.tCompound;
                    column = updateVariables(resultSet, comp,
                                             i, i+nComponents, column-1);
                    SqlWarningFailure.check(resultSet);
                    i += nComponents;
                }
            }
        }
        /*
	} catch(IllegalAccessException ex) {
            //throw new IllegalAccessError();  //2001-02-21  Knabe
            throw new Failure(ex);

        }*/
	return column;
    }


    /**Checks, that the two objects taken from this table are equal in each
        attribute value.
        @author Christoph Knabe, 2001-05-24
        @throws NotEqualExc If the two objects differ in at most one attribute
            value.
    */
    public void checkEquals(final Object i_object1, final Object i_object2)
    throws NotEqualExc, Failure
    {
        try {
            for(int i=0; i<fields.length; i++){
                final FieldDescriptor fd = fields[i];
                final Object value1 = fd.field.get(i_object1);
                final Object value2 = fd.field.get(i_object2);
                if(!value1.equals(value2)){throw new NotEqualExc(
                    fd.name, value1, value2
                );}
            }
        } catch(Exception ex){throw new Failure(ex);}
    }


    /**Copies the field values of i_object1 to o_object2.
       Requires that the two objects are taken from this table.
        @author Christoph Knabe, 2001-05-26
    */
    public void copy(final Object i_object1, final Object o_object2)
    throws Failure
    {
        try {
            for(int i=0; i<fields.length; i++){
                final FieldDescriptor fd = fields[i];
                final Object value1 = fd.field.get(i_object1);
                fd.field.set(o_object2, value1);
            }
        } catch(Exception ex){throw new Failure(ex);}
    }

    /** Unable to start up the JORA Table class.
        @author Christoph Knabe 2001-02-27
    */
    public static class StartupFailure extends Failure {

        public StartupFailure(final Exception i_cause){
            super("Unable to start up the JORA Table class", i_cause);
        }

    }

    //2001-04-19  Christoph Knabe  With unnamed exception parameters
    /** Unable to initialize a JORA Table object.
        @author Christoph Knabe 2001-02-27
    */
    public static class InitFailure extends Failure {
        public InitFailure(
            final Exception i_cause, final String i_className,
            final String i_tableName, final String i_keys
        ){
            super(
                "Unable to initialize Table {0} "
                + "for class {1} with keys {2}", i_cause, i_tableName, i_className, i_keys
            );
        }

    }

    //2001-04-19  Christoph Knabe  With unnamed exception parameters
    //2001-02-27  Knabe  Renamed from ...Error to ...Failure, as MulTEx exception,
    //                   as inner class moved to here
    /** Exception raised on initialization or when update/remove operation is
     *  invoked for a table
     *  with no correct primary key defined (key was not specified or
     *  type of the key component is not atomic).
      @author Christoph Knabe 2001-02-27
     */
    public class NoPrimaryKeyFailure extends jora.Failure {
        NoPrimaryKeyFailure(final String i_keyName) {
            super("Primary Key {0} is not atomic (i.e. missing or compound)",
               (Exception)null, i_keyName
            );
        }
    }


    /** Unable to start up the JORA Table class.
        @author Christoph Knabe 2001-02-27
    */
    public class NotEqualExc extends multex.Exc {

        public NotEqualExc(final String i_fieldName,
            final Object i_value1, final Object i_value2
        ){
            super("The 2 objects of type {0} differ in field \"{1}\" "
                + "by values {2} vs. {3}",
                cls.getName(), i_fieldName, i_value1,  i_value2
            );
        }

    }

}//Table